import os
import itkdb
import jwt
import pandas as pd
from datetime import datetime, timedelta
from fastapi import FastAPI, HTTPException, Depends
from fastapi.responses import RedirectResponse
from fastapi.security import APIKeyHeader
from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from .ReadSpec import getSpec, getSpecList

api_key_header = APIKeyHeader(name="Authorization", auto_error=False)

app = FastAPI()

# Configuration from environment variables
SECRET_KEY = os.getenv("SECRET_KEY")
ALGORITHM = os.getenv("ALGORITHM")
TOKEN_DURATION = int(os.getenv("TOKEN_DURATION"))


# Model for authentication input
class AuthInput(BaseModel):
    access_code1: str = Field(default=None,examples=["access_code1"])
    access_code2: str = Field(default=None,examples=["access_code2"])

# Model for authentication output
class AuthOutput(BaseModel):
    access_token: str = Field(default=None,examples=["access_token"])
    token_type: str = Field(default="bearer",examples=["bearer"])


# Model for the input of parameters
class SpecDataInput(BaseModel):
    project: str = Field(default=None,examples=["P"])
    componentType: str = Field(default=None,examples=["PCB"])
    testType: str = Field(default=None,examples=["METROLOGY"])
    stage: str = Field(default=None,examples=["PCB_RECEPTION"])
    parameter: Optional[str] = Field(default=None,examples=["BOW1"])

# Model for the output of parameters
class SpecDataOutput(BaseModel):
    project: Optional[str] = None
    componentType: Optional[str] = None
    testType: Optional[str] = None
    stage: Optional[str] = None
    parameter: List[str] = None
    specList: Dict[str, Dict[str, Any]] = None
    

#################
### Functions ###
#################


def Authenticate(access_code1: str, access_code2: str):
    user = itkdb.core.User(access_code1=access_code1, access_code2=access_code2)
    try:
        user.authenticate()
    except Exception as e:
        print(e)
        raise HTTPException(status_code=401, detail="Authentication failed")
        return None
    return access_code1


def Create_Token(access_code1: str):
    expire = datetime.utcnow() + timedelta(minutes=TOKEN_DURATION)
    payload = {
        "access_code1": access_code1,
        "exp": expire
    }
    
    token = jwt.encode(payload, SECRET_KEY, algorithm=ALGORITHM)
    return token


def get_api_key(authorization: str = Depends(api_key_header)):
    # Check if token is present in header (Authorization = Bearer <token>)
    if authorization is None:
        raise HTTPException(status_code=403, detail="No token")
    return authorization


#################
### ENDPOINTS ###
#################


### Redirect to swagger
@app.get("/", include_in_schema=False)
async def redirect_to_docs():
    return RedirectResponse(url="/docs")


### Authentication
@app.post("/auth")
async def auth(kwargs: AuthInput):
    kwargs_dict=kwargs.dict()

    # authenticate
    Authenticate(kwargs_dict["access_code1"], kwargs_dict["access_code2"])

    # create token
    token = Create_Token(kwargs_dict["access_code1"])

    # output structure
    output = AuthOutput(
        access_token=token,
        token_type="bearer"
    )

    return output


### Single specification retrieval
@app.post("/spec")
async def specRetrieval(kwargs: SpecDataInput, token: str = Depends(get_api_key)):
    kwargs_dict=kwargs.dict()
    
    # gatekeeping
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        ac1: str = payload.get("access_code1")
        if ac1 is None:
            raise HTTPException(status_code=401, detail="Invalid token")
    except Exception as e:
        print(e)
        raise HTTPException(status_code=401, detail="Authentication failed")
    
    # function
    query = getSpec(**kwargs_dict)

    # output structure
    output = SpecDataOutput(
        project=kwargs_dict["project"],
        componentType=kwargs_dict["componentType"],
        testType=kwargs_dict["testType"],
        stage=kwargs_dict["stage"],
        parameter=[query["parameter"]],
        specList={query["parameter"]:query["spec"]}
    )

    return output


### Multiple specifications retrieval
@app.post("/speclist")
async def specListRetrieval(kwargs: SpecDataInput, token: str = Depends(get_api_key)):
    kwargs_dict=kwargs.dict()

    # gatekeeping
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        ac1: str = payload.get("access_code1")
        if ac1 is None:
            raise HTTPException(status_code=401, detail="Invalid token")
    except Exception as e:
        print(e)
        raise HTTPException(status_code=401, detail="Invalid token")

    # function
    query = getSpecList(**kwargs_dict)

    # output structure
    output = SpecDataOutput(
        project=kwargs_dict["project"],
        componentType=kwargs_dict["componentType"],
        testType=kwargs_dict["testType"],
        stage=kwargs_dict["stage"],
        parameter=query["parameter"],
        specList={x : y | {"associatedParam": []} for x, y in zip(query["parameter"], query["spec"])}
    )
    return output


# Example endpoint to read modules
@app.get("/modules")
def read_modules():
    csv_folder_path = "/spec_files"

    if not os.path.exists(csv_folder_path):
        raise HTTPException(status_code=404, detail="CSV files not found")
    try:
        df = pd.read_csv(csv_folder_path)
        return df.to_dict(orient="records")
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Example endpoint to read a specific module by ID
@app.get("/modules/{module_id}")
def read_module(module_id: int):
    # treating the module id input
    if module_id[-4:] != ".csv":
        module_id = module_id + ".csv"

    # csv file path if the working directory is the root of the project
    csv_file_path = f"/spec_files/{module_id}"

    if not os.path.exists(csv_file_path):
        raise HTTPException(status_code=404, detail="CSV file not found")
    try:
        df = pd.read_csv(csv_file_path)
        if module_id >= len(df) or module_id < 0:
            raise HTTPException(status_code=404, detail="Module ID not found")
        return df.iloc[module_id].to_dict()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


### Health check
@app.get("/health")
async def health_check():
    return {"status": "ok"}


### Help
@app.get("/help")
def help():
    return {
        "message": "Welcome to itk-spec",
        "endpoints": {
            "/spec": {
                "method": "POST",
                "description": "Get specification for a single parameter",
                "parameters": {
                    "project": "Project name",
                    "componentType": "Component type",
                    "testType": "Test type",
                    "stage": "Stage",
                    "parameter": "Parameter"
                }
            },
            "/speclist": {
                "method": "POST",
                "description": "Get specifications for multiple parameters",
                "parameters": {
                    "project": "Project name",
                    "componentType": "Component type",
                    "testType": "Test type",
                    "stage": "Stage",
                }
            },
            "/health": {
                "method": "GET",
                "description": "Check for API server status",
            },
        }
    }

if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000, log_level="info")


import itkdb
from fastapi import HTTPException
from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
import json
import requests
import csv
from io import StringIO
import os
from datetime import datetime, timedelta

__version__ = "1.5.0"

# Configuration - you can override these by setting environment variables
DEFAULT_API_BASE_URL = "https://key-version-tester.app.cern.ch"

############
## Authentication token for creating ITKSpecClient
############

class AuthToken:
    """Authentication token object required to create ITKSpecClient"""
    
    def __init__(self, token: str, expiry: datetime, access_code1: str, access_code2: str):
        """
        Internal constructor - should not be called directly
        Use authenticate() function to get an AuthToken
        """
        self._token = token
        self._expiry = expiry
        self._access_code1 = access_code1
        self._access_code2 = access_code2
    
    def is_valid(self) -> bool:
        """Check if the authentication token is still valid"""
        return datetime.now() < self._expiry
    
    def time_remaining(self) -> timedelta:
        """Get the remaining time before the token expires"""
        if not self.is_valid():
            return timedelta(0)
        return self._expiry - datetime.now()

def authenticate(access_code1: str, access_code2: str) -> AuthToken:
    """
    Authenticate with ITKDB and return an AuthToken required to create clients
    
    Args:
        access_code1: First access code for ITKDB authentication
        access_code2: Second access code for ITKDB authentication
    
    Returns:
        AuthToken object required to create ITKSpecClient
        
    Raises:
        Exception: If authentication fails
        
    Example:
        auth_token = itkspec.authenticate('access_code1', 'access_code2')
        client = itkspec.ITKSpecClient(auth_token)
    """
    api_base_url = DEFAULT_API_BASE_URL
    auth_url = f"{api_base_url}/auth"
    auth_data = {
        "access_code1": access_code1,
        "access_code2": access_code2
    }
    
    try:
        response = requests.post(auth_url, json=auth_data)
        response.raise_for_status()
        
        auth_response = response.json()
        token = auth_response["access_token"]
        expiry = datetime.now() + timedelta(minutes=60)  # Default expiry time of 60 minutes
        
        return AuthToken(token, expiry, access_code1, access_code2)
        
    except requests.RequestException as e:
        raise Exception(f"Authentication failed: {str(e)}")

class ITKSpecClient:
    """Main client class for ITK Specifications API"""
    
    def __init__(self, auth_token: AuthToken):
        """
        Initialize the ITK Spec client with an AuthToken
        
        Args:
            auth_token: AuthToken object obtained from authenticate() function
        
        Raises:
            Exception: If auth_token is invalid or expired
            
        Example:
            auth_token = itkspec.authenticate('access_code1', 'access_code2')
            client = itkspec.ITKSpecClient(auth_token)
        """
        if not isinstance(auth_token, AuthToken):
            raise TypeError("Client requires an AuthToken. Use authenticate() function first.")
        
        if not auth_token.is_valid():
            raise Exception("AuthToken has expired. Please authenticate again.")
        
        self.auth_token = auth_token
        self.api_base_url = DEFAULT_API_BASE_URL
    
    # checking jwt token
    def time_remaining(self) -> timedelta:
        """Get the remaining time before the client expires"""
        return self.auth_token.time_remaining()
    
    # checking jwt token
    def _check_token_validity(self) -> None:
        """Check if the token is still valid, raise exception if expired"""
        if not self.auth_token.is_valid():
            raise Exception("Client session has expired. Please authenticate again.")
    
    def _get_token(self) -> str:
        """Get the authentication token"""
        self._check_token_validity()
        return self.auth_token._token
    
    def _make_authenticated_request(self, endpoint: str, data: Dict = None, method: str = "POST") -> Dict:
        """Make an authenticated request to the API"""
        token = self._get_token()
        headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json"
        }
        
        url = f"{self.api_base_url}{endpoint}"
        
        try:
            if method.upper() == "POST" and data:
                response = requests.post(url, json=data, headers=headers)
            elif method.upper() == "GET":
                response = requests.get(url, headers=headers)
            else:
                response = requests.request(method, url, json=data, headers=headers)
                
            response.raise_for_status()
            return response.json()
            
        except requests.RequestException as e:
            raise Exception(f"API request failed: {str(e)}")

    def retrieve_spec(self, data: Dict) -> Dict:
        """
        Get specification for a single parameter
        
        Args:
            data: Dict with spec parameters containing: 
                  project, componentType, testType, stage, parameter
            
        Returns:
            Dict containing the specification data
            
        Example:
            client.retrieve_spec({'project': 'P', 'componentType': 'PCB', 
                                 'testType': 'METROLOGY', 'stage': 'PCB_RECEPTION', 
                                 'parameter': 'BOW1'})
        """
        required_fields = ['project', 'componentType', 'testType', 'stage', 'parameter']
        missing_fields = [field for field in required_fields if field not in data]
        
        if missing_fields:
            raise ValueError(f"Missing required fields in data: {missing_fields}")
        
        return self._make_authenticated_request("/spec", data)

    def retrieve_spec_list(self, data: Dict) -> Dict:
        """
        Get specifications for multiple parameters
        
        Args:
            data: Dict with spec parameters containing: 
                  project, componentType, testType, stage
            
        Returns:
            Dict containing the specification list data
            
        Example:
            client.retrieve_spec_list({'project': 'P', 'componentType': 'PCB', 
                                      'testType': 'METROLOGY', 'stage': 'PCB_RECEPTION'})
        """
        required_fields = ['project', 'componentType', 'testType', 'stage']
        missing_fields = [field for field in required_fields if field not in data]
        
        if missing_fields:
            raise ValueError(f"Missing required fields in data: {missing_fields}")
        
        return self._make_authenticated_request("/speclist", data)
    
    def health_check(self) -> Dict:
        """Check API server status"""
        return self._make_authenticated_request("/health", method="GET")

#################
### OLD FUNCTIONS - Updated to use ITKSpecClient ###
#################

#### authenticate user - now returns ITKSpecClient using new pattern
def authenticate_user(accessCode1: str, accessCode2: str):
    """Legacy function - authenticates and returns a client"""
    auth_token = authenticate(accessCode1, accessCode2)
    return ITKSpecClient(auth_token)

### get single spec - updated to use token-based authentication
def retrieve_spec(client_or_ACs, data: Dict[str, str] = None):
    """
    Get specification for a single parameter - supports both new and legacy usage
    
    Args:
        client_or_ACs: Either ITKSpecClient instance or dict with access codes
        data: Dict with spec parameters (project, componentType, testType, stage, parameter)
    
    Returns:
        Dict containing the specification data
    """
    if isinstance(client_or_ACs, ITKSpecClient):
        # New client-based usage
        client = client_or_ACs
        if data:
            return client.retrieve_spec(data)
        else:
            raise ValueError("data parameter is required")
    else:
        # Legacy usage - create client from access codes using new pattern
        ACs = client_or_ACs
        auth_token = authenticate(ACs['accessCode1'], ACs['accessCode2'])
        client = ITKSpecClient(auth_token)
        if data:
            return client.retrieve_spec(data)
        else:
            raise ValueError("data parameter is required")

### get multiple specs - updated to use token-based authentication
def retrieve_specList(client_or_ACs, data: Dict[str, str] = None):
    """
    Get specifications for multiple parameters - supports both new and legacy usage
    
    Args:
        client_or_ACs: Either ITKSpecClient instance or dict with access codes
        data: Dict with spec parameters (project, componentType, testType, stage)
    
    Returns:
        Dict containing the specification list data
    """
    if isinstance(client_or_ACs, ITKSpecClient):
        # New client-based usage
        client = client_or_ACs
        if data:
            return client.retrieve_spec_list(data)
        else:
            raise ValueError("data parameter is required")
    else:
        # Legacy usage - create client from access codes using new pattern
        ACs = client_or_ACs
        auth_token = authenticate(ACs['accessCode1'], ACs['accessCode2'])
        client = ITKSpecClient(auth_token)
        if data:
            return client.retrieve_spec_list(data)
        else:
            raise ValueError("data parameter is required")

### api help - updated to use new API
def api_help():
    """Get API help from the server"""
    base_url = DEFAULT_API_BASE_URL
    try:
        url = f"{base_url}/help"
        response = requests.get(url)
        response.raise_for_status()
        result = response.json()
        print(json.dumps(result, indent=2))
        return result
    except requests.RequestException as e:
        print(f"Failed to get API help: {e}")
        return None

### server check - updated to use new API
def health_check():
    """Check API server status"""
    base_url = DEFAULT_API_BASE_URL
    try:
        url = f"{base_url}/health"
        response = requests.get(url)
        response.raise_for_status()
        result = response.json()
        print(json.dumps(result, indent=2))
        return result
    except requests.RequestException as e:
        print(f"Failed to check server health: {e}")
        return None

### help
def help():
    help_text = """ITK Specifications Package Functions:

NEW RECOMMENDED USAGE:
    # Step 1: Authenticate first to get an AuthToken
    try:
        auth_token = itkspec.authenticate('your_access_code1', 'your_access_code2')
        print(f"Authentication successful. Token valid for: {auth_token.time_remaining()}")
    except Exception as e:
        print(f"Authentication failed: {e}")
        return
    
    # Step 2: Create client with the AuthToken
    try:
        client = itkspec.ITKSpecClient(auth_token)
        print(f"Client created successfully. Valid for: {client.time_remaining()}")
    except Exception as e:
        print(f"Client creation failed: {e}")
        return
    
    # Step 3: Use the client (check validity if needed)
    if client.is_valid():
        # Get single specification using data dictionary
        spec = client.retrieve_spec({'project': 'P', 'componentType': 'PCB', 
                                    'testType': 'METROLOGY', 'stage': 'PCB_RECEPTION', 
                                    'parameter': 'BOW1'})
        print(spec)
        
        # Get multiple specifications using data dictionary
        specs = client.retrieve_spec_list({'project': 'P', 'componentType': 'PCB',
                                          'testType': 'METROLOGY', 'stage': 'PCB_RECEPTION'})
        print(specs)
        
        # Check server health
        status = client.health_check()
        print(status)
    else:
        print("Client session expired. Authenticate again.")

ALTERNATIVE SINGLE-STEP USAGE:
    # Authenticate and create client in one step (legacy-style)
    client = itkspec.authenticate_user('access_code1', 'access_code2')

LEGACY FUNCTIONS (still supported):
    1. authenticate - Get an AuthToken for creating clients
    2. authenticate_user - returns a client if the user is verified
    3. retrieve_spec - Get specification for a single parameter
    4. retrieve_specList - Get specifications for multiple parameters
    5. api_help - List the available API endpoints
    6. health_check - Check for API server status
    7. get_from_EOS - Get backup data from EOS
    8. Exit
    """
    print(help_text)

    while True:
        choice = input("Enter the number of the function you would like an example of (or type 'exit' to quit): ").strip().lower()
        if choice in ["8", "exit"]:
            print("Exiting help menu.")
            break
        details = {
            "1": "Example: auth_token = itkspec.authenticate('your_access_code1', 'your_access_code2')",
            "2": "Example: client = itkspec.authenticate_user('your_access_code1', 'your_access_code2')",
            "3": "Example: spec = itkspec.retrieve_spec(client, {'project':'P', 'componentType':'PCB', 'testType':'METROLOGY', 'stage':'PCB_RECEPTION', 'parameter':'BOW1'})",
            "4": "Example: specs = itkspec.retrieve_specList(client, {'project':'P', 'componentType':'PCB', 'testType':'METROLOGY', 'stage':'PCB_RECEPTION'})",
            "5": "Example: itkspec.api_help()",
            "6": "Example: itkspec.health_check()",
            "7": "Example: data = itkspec.get_from_EOS('filename.csv')",
        }
        print(details.get(choice, "Invalid selection. Please enter a number from 1 to 8."))

### data backup from EOS
def get_from_EOS(filename):
    url = f'https://eddossan.web.cern.ch/itk-specs/{filename}'
    reqs = requests.get(url)
    if reqs.status_code != 200:
        print(f"Failed to retrieve the URL: {url}")
        return []
    csv_content = reqs.text
    print(f"Fetched content from {url}:\n{csv_content[:500]}...")

    # Parse the CSV content
    data = []
    csv_reader = csv.reader(StringIO(csv_content))
    headers = ["project", "componentType", "testType", "stage", "parameter", "spec"]
    for row in csv_reader:
        if row:
            row_data = {headers[i]: row[i] for i in range(len(headers))}
            if row_data["spec"]:
                try:
                    row_data["spec"] = json.loads(row_data["spec"].replace("'", "\""))  # Convert spec string to dictionary
                except json.JSONDecodeError:
                    print(f"Failed to decode spec: {row_data['spec']}")
                    row_data["spec"] = {}
            else:
                row_data["spec"] = {}
            data.append(row_data)

    print(f"Data: {data}")
    return data


### imports
import os.path
import os
import pandas as pd
from ast import literal_eval


### template info. for user input
infoDict_template={'project':None, 'componentType':None, 'testType':None, 'stage':None, 'parameter':None}
## copy function - for use in functions
def copyTemplate():
    return infoDict_template.copy()


### Function for returning a single spec from a unique parameter
def getSpec(**kwargs):
    inCheck = True
    core_keys = ['project', 'componentType', 'testType', 'stage', 'parameter']

    # Check for missing core labels
    if any(kwargs.get(label) is None for label in core_keys):
        print(" there's a core label missing. Please check inputs.")
        inCheck = False

    # Determine the file name
    fileName = "_".join([kwargs['project'], kwargs['componentType']]) + ".csv"

    # Check if the file exists
    if fileName is not None:
        # Get root directory
        current_directory = os.path.dirname(os.path.abspath(__file__))

        # Construct the full path to the spec_files directory
        spec_files_directory = os.path.join(current_directory, "spec_files")
        print(f"  csv files directory: {spec_files_directory}")

        # Construct the full path to the file
        file_path = os.path.join(spec_files_directory, fileName)
        try:
            os.path.isfile(file_path)
            print(f"  file path: {file_path}\n")
        except FileNotFoundError:
            print("Error: This specification file does not exist yet or inputs are incorrect.")
            inCheck = False

    if not inCheck:
        print("returning nothing :(")
        return pd.DataFrame()

    # Open the file and get spec data
    print("# getting spec from csv #\n")
    df_csv = pd.read_csv(file_path, converters={"spec": literal_eval})

    queryStr=f'project==\"{kwargs["project"]}\" & componentType==\"{kwargs["componentType"]}\" & parameter==\"{kwargs["parameter"]}\"'
    print(f"query spec using: \n{queryStr}\n")

    df_spec=df_csv.query(queryStr)

    if df_spec.empty:
        print(" - empty spec csv :(.")
        return pd.DataFrame()
    else:
        spec_dict = df_spec.to_dict()
        out_dict = {
            'parameter': spec_dict['parameter'][0],
            'spec': spec_dict['spec'][0]
        }
        print(df_spec.to_markdown())
    return out_dict


### Function for returning multiple specs from various parameters
def getSpecList(**kwargs):
    inCheck = True
    core_keys = ['project', 'componentType', 'testType', 'stage']

    # Check for missing core labels
    if any(kwargs.get(label) is None for label in core_keys):
        print(" there's a core label missing. Please check inputs.")
        inCheck = False

    # Determine the file name
    fileName = "_".join([kwargs['project'], kwargs['componentType']]) + ".csv"

    # Check if the file exists
    if fileName is not None:
        # Get the main directory
        current_directory = os.path.dirname(os.path.abspath(__file__))

        # Construct the full path to the spec_files directory
        spec_files_directory = os.path.join(current_directory, "spec_files")
        print(f"  csv files directory: {spec_files_directory}")

        # Construct the full path to the file
        file_path = os.path.join(spec_files_directory, fileName)
        try:
            os.path.isfile(file_path)
            print(f"  file path: {file_path}")
        except FileNotFoundError:
            print("Error: This specification file does not exist yet or inputs are incorrect.")
            inCheck = False

    if not inCheck:
        print("returning nothing :(")
        return pd.DataFrame()

    # Open the file and get spec data
    print("# getting spec from csv #\n")
    df_csv = pd.read_csv(file_path, converters={"spec": literal_eval})

    queryStr=f'project==\"{kwargs["project"]}\" & componentType==\"{kwargs["componentType"]}\"'
    print(f"query spec using: \n{queryStr}\n")
    
    df_spec=df_csv.query(queryStr)

    if df_spec.empty:
        print(" - empty spec csv :(.")
        return pd.DataFrame()
    else:
        spec_dict = df_spec.to_dict()
        out_dict = {
            'parameter': list(spec_dict['parameter'].values()),
            'spec': list(spec_dict['spec'].values())
        }
        print(df_spec.to_markdown())
        print(out_dict)
    return out_dict


def listSpecFiles():
    current_directory = os.path.dirname(os.path.abspath(__file__))
    spec_files_directory = os.path.join(current_directory, "spec_files")
    
    ## get list of (csv) files in directory
    csv_list=os.listdir(spec_files_directory)
    ## filter csv files
    csv_list=[x for x in csv_list if ".csv" in x]
    
    print(f"csv files in directory: {csv_list}")
    return csv_list